package aa.display;

import java.awt.*;
import java.awt.event.*;
import aa.node.Heuristic;
import aa.node.SearchNode;
import aa.mynode.FifteenNode;
import javax.swing.*;

/**
 * This panel implements the display of the current node state.
 * It also allows a user to click on a block in order to move the
 * block to the free space.
 * @author: Darius Cooper
 */
public class FifteenPanel extends javax.swing.JPanel {

	private static final Color COLOR_BUTTONS = new Color(60, 165, 170);
	private static final Color COLOR_FREE = new Color(170,180,254);

	/** 
	 * The time (in milliseconds) to wait when a incorrect operation
	 * is attempted. Waiting gives the user visual feedback, showing the
	 * button being depressed. This is simple, non-intrusive feedback.
	 */ 
	private static final int MILLISECONDS_INVALID_OPERATION = 300;
	
	//The object that is listening for user moves to be performed
	ValidUserMoveListener validUserMoveListener = null;

	//The <code>FifteenNode</code> object being displayed by panel
	FifteenNode nodeState = null;
	//The buttons that fill the grid
	private javax.swing.JButton[][] buttons = new JButton[FifteenNode.GRID_SIZE][FifteenNode.GRID_SIZE];

	//Indicates that the user should not be able to change the state (it is
	//a "display only" state. However, programs may still change the state
	//via programming.
	private boolean displayOnly = false;
/**
 * Default constructor. Sets up the buttons without any specific
 * numbers, because there is no <code>NodeState</code> to represent.
 */
public FifteenPanel() {

	this.setLayout(
		new java.awt.GridLayout(FifteenNode.GRID_SIZE, FifteenNode.GRID_SIZE));
	for (int i = 0; i < FifteenNode.GRID_SIZE; i++) {
		for (int j = 0; j < FifteenNode.GRID_SIZE; j++) {
			buttons[i][j] = (new JButton());
			buttons[i][j].setFont(new Font("Arial",Font.BOLD,32));
			buttons[i][j].addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {

					handleButtonClick((JButton) evt.getSource());
				}
			});

			this.add(buttons[i][j]);
		}
	}
}
/**
 * Constructor to set up the panel and display the <code>FifteenNode</code> that
 * is provided.
 * @param nodeState The <code>FifteenNode</code> node state to be displayed.
 */
public FifteenPanel(FifteenNode nodeState) {

	this();
	this.nodeState = nodeState;
	update(nodeState);
}
/**
 * Adds a <code>ValidUserMoveListener</code> to listen for user moves.
 * @param listener aa.display.ValidMoveListener
 */
public void addValidUserMoveListener(ValidUserMoveListener listener) {

	this.validUserMoveListener = listener;
}
/**
 * Gets the <code>FifteenNode</code> currently being displayed by the panel.
 * @return The <code>FifteenNode</code> that is currently being displayed
 * by the panel. 
 */
public FifteenNode getNodeState() {
	return nodeState;
}
/**
 * Handles the <code>ActionEvent</code> of a user clicking on a button
 * in the panel, in order to move a block. If the move was a valid one, this
 * method will also trigger a <code>ValidUserMoveEvent</code>.
 *
 * @param clickedButton The button that was clicked.
 */
public void handleButtonClick(JButton clickedButton) {

	//If the panel is currently set to "Display-only" mode, ignore the click
	if (isDisplayOnly()) {
		return;
	}

	//Identify the grid-location of the clicked button
	//To-do: This is something that can be optimized in a later release
	for (int i = 0; i < FifteenNode.GRID_SIZE; i++) {
		for (int j = 0; j < FifteenNode.GRID_SIZE; j++) {
			if (buttons[i][j] == clickedButton) {
				try {
					this.nodeState.moveBlock(i, j);
					update(this.nodeState);
					if (validUserMoveListener != null) {
						validUserMoveListener.validUserMovePerformed(new ValidUserMoveEvent(this.nodeState));
					}
				} catch (aa.node.NodeState.InvalidOperationException e1) {
					try {
						Thread.sleep(MILLISECONDS_INVALID_OPERATION);
					} catch (InterruptedException e2) {
					}
				}
			}
		}
	}
}
/**
 * Checks if the panel is display-only or user modify-able
 *
 * @return <code>true</code> if display-only, otherwise <code>false</code>
 */
public boolean isDisplayOnly() {
	return displayOnly;
}


public static void main(String[] args) {

	FifteenNode testState = new FifteenNode();
	SearchNode startNode = new SearchNode(testState);
	Heuristic heuristic = new aa.mynode.FifteenHeuristic();

	
	
	FifteenPanel testPanel = new FifteenPanel(testState);
	


	JFrame frame = new JFrame();		
	frame.getContentPane().add(testPanel);
	frame.setSize(300,300);
	frame.setVisible(true);
	frame.show();
}
/**
 * Sets the panel to be display-only or user modify-able
 *
 * @param newDisplayOnly <code>true</code> will set display-only, while
 * <code>false</code> will set the panel to user modify-able.
 */
public void setDisplayOnly(boolean newDisplayOnly) {
	displayOnly = newDisplayOnly;
}
/**
 * Updates a new state to be displayed in the panel.
 *
 * @param nodeState The new <code>FifteenNode</code> object to be dislayed
 * in the panel.
 */
public void update(FifteenNode nodeState) {

	this.nodeState = nodeState;
	for (int i = 0; i < FifteenNode.GRID_SIZE; i++) {
		for (int j = 0; j < FifteenNode.GRID_SIZE; j++) {
			if (nodeState.isFreeBlock(i,j)) {
				buttons[i][j].setText("");
				buttons[i][j].setBackground(COLOR_FREE);
			} else {
				buttons[i][j].setText(Integer.toString(nodeState.getBlockNum(i, j)));
				buttons[i][j].setBackground(COLOR_BUTTONS);

			}
		}

	}
	invalidate();
	repaint();
}
}