package aa.node;

import java.util.TreeSet;
import java.util.HashMap;
/**
 * This class performs the search for a goal-state. 
 * Since searching can take time, it extends Thread and can therefore 
 * easily be run as a separate thread.
 */


public class Searcher extends Thread {
	/* An array of search nodes, 
	   position zero has the start node
	   the last position has the solution node
	 */
	private aa.node.SearchNode[] solutionNodes;
	//A copy of the start node that is passed in by the caller
	private SearchNode startNode;
	private Heuristic heuristic; //The heuristic to be used in the search
	//This flag indicates if a caller from another thread has requested
	//that the search be terminated
	private boolean stopRequested = false;
	private boolean debugTrace = false; //True means: write trace-info to console
	private boolean solveRequested = false; //Used as a semaphore to show if a solve is going on
/**
 * Searcher constructor
 * @param startNode Any valid <code>SearchNode</code> that should be used as a starting point
 * @param heuristic A heuristic that is valid for the type of startNode that
 * has been provided. Note that the solutionState is not provided as that is
 * encapsulated by the heuristic.
 * The search calls <code>Thread.yield()</code> at every node, so that other threads
 * remain responsive. At each such point, it also checks if some other thread has
 * requested the search to stop.
 */
public Searcher(SearchNode startNode, Heuristic heuristic) {
	super();
	this.startNode = startNode;
	this.heuristic = heuristic;
}
/**
 * Returns an array of <code>SearchNode</code> objects. 
 * The zero-th position contains the start node, the last position contains 
 * the solution node. 
 * The positions in between are the states that lead from start-node to solution node.
 * @return aa.node.SearchNode[] containing a solution, or
 *         <code>null</code> is the solution has not yet been found.
 */
public aa.node.SearchNode[] getSolutionNodes() {
	return solutionNodes;
}
/**
 * Check if debug-tracing is on.
 * @return <code>true</code> is tracing is on, else return <code>false</code>.
 */
public boolean isDebugTrace() {
	return debugTrace;
}
/**
 * Checks if the puzzle is currently in the process of being solved
 * @return <code>true</code> is being solved, else <code>false</code>
 */
synchronized public boolean isSolving() {
	return solveRequested;
}
/**
 * Checks if the thread has received a request to stop its search for a solution.
 * @return <code>true</code> is a stop has been requested, else return <code>false<code>.
 */
synchronized public boolean isStopRequested() {
	return stopRequested;
}
/**
 * When an object implementing interface <code>Runnable</code> is used 
 * to create a thread, starting the thread causes the object's 
 * <code>run</code> method to be called in that separately executing 
 * thread. 
 * <p>
 * The general contract of the method <code>run</code> is that it may 
 * take any action whatsoever.
 *
 * @see     java.lang.Thread#run()
 */
public void run() {

	setSolveRequested(true);

	for (;;) {
		if (isSolving()) {
			setStopRequested(false);
			search();
			System.out.println("Searcher search() method ending");
			setSolveRequested(false);
		}
		try {
			sleep(2000);
		} catch (InterruptedException ex) {
			//}

		}
	}
}
/**
 * Searches for an optimal solution from the start state to the solution.
 * If debug-trace has been set to <code>true</code>, then each node that is
 * examined is written using <code>System.out.println()</code> to allow for
 * debugging.
 * When this method completes successfully, it will update the solution-set
 * array, which can then be accessed by a call to <code>getSolutionSet()</code>
 * 
 */
public void search() {

	//Create a new copy a start node, using only the state from the input
	startNode = new SearchNode(startNode.getNodeState());

	solutionNodes = null;
	int currentCost = 0;

	HashMap stateMap = new HashMap();
	TreeSet nodeList = new TreeSet();
	int nodeListMin = Integer.MAX_VALUE;

	SearchNode currentNode = startNode;
	//Place Start node in the Pool
	stateMap.put(currentNode.getNodeState(), currentNode);
	nodeList.add(currentNode);

	int nodeNumber = 0;
	while (!currentNode.isGoalState()) {
		Thread.yield();
		if (isStopRequested()) {
			solutionNodes = null;
			return;
		}
		nodeNumber++;
		if (this.isDebugTrace()) {
			System.out.println("Current " + currentNode);
		}
		if (currentNode.isExpanded()) {
			solutionNodes = null;
			System.out.println("Cannot find a solution!");
			break;
		}
		nodeList.remove(currentNode);
		//currentNode.setExpanded(true);
		SearchNode[] expandedNodes = currentNode.getNextNodeList();
		//int expandedListMin = Integer.MAX_VALUE;
		SearchNode expandedMinNode = null;

		for (int i = 0; i < expandedNodes.length; i++) {
			if (expandedNodes[i] == null) {
				break; //Handles situation where the returned array 
				//is not completely filled
			}
			Thread.yield();
			if (this.isDebugTrace()) {
				System.out.println(
					"                            Expanded # " + i + "= " + expandedNodes[i]);
			}
			NodeState currentNodeState = expandedNodes[i].getNodeState();
			SearchNode existingSearchNode = (SearchNode) stateMap.get(currentNodeState);
			if (existingSearchNode != null) {
				if (existingSearchNode.getLowestCost() <= expandedNodes[i].getLowestCost()) {
					//Ignore the current route to this node if we already 
					//visited this node and have not
					//found a lower-cost route to reach this node
					continue;
				} else {
					//No need to recalculate the heuristic
					expandedNodes[i].setEstimateToGoal(existingSearchNode.getEstimateToGoal());
					//Remove the node from the pool (it will be replaced by the new one)
					stateMap.remove(currentNodeState);
					nodeList.remove(existingSearchNode);
					//What else? Change anything about the costs ...back-tracking?
				}
			} else {
				//Calculate the heuistic for the newly expanded node
				expandedNodes[i].setEstimateToGoal(
					heuristic.getEstimatedCostToGoal(expandedNodes[i]));
			}
			//Calc Tot Est
			int currentTotEst =
				expandedNodes[i].getLowestCost() + expandedNodes[i].getEstimateToGoal();
			if (currentTotEst < nodeListMin) {
				nodeListMin = currentTotEst;
				expandedMinNode = expandedNodes[i];
			}
			//Place in Pool
			stateMap.put(currentNodeState, expandedNodes[i]);
			nodeList.add(expandedNodes[i]);

		}
		if (expandedMinNode != null) { //Found something lower than the pool
			currentNode = expandedMinNode;
		} else {
			//Backtrack..re-setting the costs
			while (currentNode.getPrevSearchNode() != null) {
				currentNode.getPrevSearchNode().setEstimateToGoal(
					currentNode.getLowestCost()
						- currentNode.getPrevSearchNode().getLowestCost()
						+ currentNode.getEstimateToGoal());
				currentNode = currentNode.getPrevSearchNode();
			}
			//Select the minimum from the pool
			currentNode = (SearchNode) nodeList.first();
		}
		//break;
	}

	//Backtrack.. to Print result
	System.out.println("========================================================");
	solutionNodes = new SearchNode[currentNode.getLowestCost() + 1];
	for (int i = solutionNodes.length - 1; currentNode != null;) {
		if (this.isDebugTrace()) {

			System.out.println(currentNode.toString() + i);
		}
		solutionNodes[i] = currentNode;
		i--;
		currentNode = currentNode.getPrevSearchNode();
	}

}
/**
 * Sets debug tracing on or off.
 * @param newDebugTrace Set to <code>true</code> to start tracing, or
 * set to <code>false</code> to turn it off.
 */
public void setDebugTrace(boolean newDebugTrace) {
	debugTrace = newDebugTrace;
}
/**
 * Sets the solveRequested semaphore
 * @param newSolving boolean
 */
synchronized private void setSolveRequested(boolean newSolveRequested) {
	solveRequested = newSolveRequested;
}
/**
 * Set a new start node to be exploer during a search.
 * 
 * @param newStartNode A new start node from where the search should begin
 */
public void setStartNode(SearchNode newStartNode) {

	this.startNode = new SearchNode(newStartNode.getNodeState());
	
}
/**
 * Request the search to stop.
 * @param newKeepRunning Set to <code>true</code> to stop the search. 
 * The search will stop the next time a new node is examined. 
 */
synchronized public void setStopRequested(boolean newStopRequested) {
	stopRequested = newStopRequested;
}
/**
 * Starts a new solution, using the start node provided in the call to
 * <code>setStartNode()</code>
 */
public void solveFromNewStartNode() {
	this.solutionNodes = null;
	setSolveRequested(true);
}
}
